#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import argparse


def build_parser():
    """
    The configuration parser for common hyperparameters of all environment.

    Prepare parameters:
        --algorithm_name <algorithm_name>
            specifiy the algorithm, including `["rmappo", "mappo", "rmappg", "mappg", "trpo"]`
        --experiment_name <str>
            an identifier to distinguish different experiment.
        --seed <int>
            set seed for numpy and torch
        --cuda
            by default True, will use GPU to train; or else will use CPU;
        --cuda_deterministic
            by default, make sure random seed effective. if set, bypass such function.
        --pytorch_threads <int>
            number of training threads working in parallel. by default 1
        --n_rollout_threads <int>
            number of parallel envs for training rollout. by default 32
        --n_eval_rollout_threads <int>
            number of parallel envs for evaluating rollout. by default 1
        --n_render_rollout_threads <int>
            number of parallel envs for rendering, could only be set as 1 for some environments.
        --num_env_steps <int>
            number of env steps to train (default: 10e6)
        --user_name <str>
            [for wandb usage], to specify user's name for simply collecting training data.
        --disable_wandb
            [for wandb usage], by default False, will log date to wandb server. or else will use tensorboard to log data.

    Env parameters:
        --env_name <str>
            specify the name of environment
        --use_obs_instead_of_state
            [only for some env] by default False, will use global state; or else will use concatenated local obs.

    Replay Buffer parameters:
        --episode_length <int>
            the max length of episode in the buffer.

    Network parameters:
        --separate_policy
            by default False, all agents will use different networks; set to make training agents with shared policies.
        --concat_obs_as_share_obs
            by default True, use centralized training mode; or else will decentralized training mode.
        --use_conv1d
            by default False, do not use conv1d. or else, will use conv1d to extract features.
        --stacked_frames <int>
            Number of input frames which should be stack together.
        --hidden_size <int>
            Dimension of hidden layers for actor/critic networks
        --layer_N <int>
            Number of layers for actor/critic networks
        --activation_id
            choose 0 to use tanh, 1 to use relu, 2 to use leaky relu, 3 to use elu
        --use_popart
            by default True, use running mean and std to normalize rewards.
        --use_feature_popart
            by default False, do not apply popart to normalize inputs. if set, apply popart to normalize inputs.
        --use_feature_normalization
            by default True, apply layernorm to normalize inputs.
        --use_orthogonal
            by default True, use Orthogonal initialization for weights and 0 initialization for biases. or else, will use xavier uniform inilialization.
        --gain
            by default 0.01, use the gain # of last action layer
        --use_naive_recurrent_policy
            by default False, use the whole trajectory to calculate hidden states.
        --use_recurrent_policy
            by default, use Recurrent Policy. If set, do not use.
        --recurrent_N <int>
            The number of recurrent layers ( default 1).
        --data_chunk_length <int>
            Time length of chunks used to train a recurrent_policy, default 10.

    Attn parameters:
        --use_attn
            by default False, use attention tactics. 
        --attn_N
            the number of attn layers, by default 1,  
        --attn_size
            by default, the hidden size of attn layer. 
        --attn_heads
            by default, the # of multiply heads. 
        --dropout
            by default 0, the dropout ratio of attn layer.  
        --use_average_pool
            by default True, use average pooling for attn model.
        --use_cat_self
            by default True, whether to strengthen own characteristics. 

    Optimizer parameters:
        --lr <float>
            learning rate parameter,  (default: 5e-4, fixed).
        --critic_lr <float>
            learning rate of critic  (default: 5e-4, fixed)
        --opti_eps <float>
            RMSprop optimizer epsilon (default: 1e-5)
        --weight_decay <float>
            coefficience of weight decay (default: 0)

    PPO parameters:
        --ppo_epoch <int>
            number of ppo epochs (default: 15)
        --use_policy_vhead
            by default, do not use policy vhead. if set, use policy vhead.
        --use_clipped_value_loss
            by default, clip loss value. If set, do not clip loss value.
        --clip_param <float>
            ppo clip parameter (default: 0.2)
        --num_mini_batch <int>
            number of batches for ppo (default: 1)
        --policy_value_loss_coef <float>
            policy value loss coefficient (default: 0.5)
        --entropy_coef <float>
            entropy term coefficient (default: 0.01)
        --use_max_grad_norm
            by default, use max norm of gradients. If set, do not use.
        --max_grad_norm <float>
            max norm of gradients (default: 0.5)
        --use_gae
            by default, use generalized advantage estimation. If set, do not use gae.
        --gamma <float>
            discount factor for rewards (default: 0.99)
        --gae_lambda <float>
            gae lambda parameter (default: 0.95)
        --use_proper_time_limits
            by default, the return value does consider limits of time. If set, compute returns with considering time limits factor.
        --use_huber_loss
            by default, use huber loss. If set, do not use huber loss.
        --use_value_active_masks
            by default True, whether to mask useless data in value loss.  
        --use_return_active_masks
            by default True, whether to mask useless data in return data.  
        --huber_delta <float>
            coefficient of huber loss.   
        --use_single_network
            by default False, whether to share base network between policy network and value network.

    PPG parameters:
        --aux_epoch <int>
            number of auxiliary epochs. (default: 4)
        --clone_coef <float>
            clone term coefficient (default: 0.01)

    Run parameters:
        --use_linear_lr_decay
            by default, do not apply linear decay to learning rate. If set, use a linear schedule on the learning rate

    Save & Log parameters:
        --save_interval <int>
            time duration between contiunous twice models saving.
        --log_interval <int>
            time duration between contiunous twice log printing.

    Eval parameters:
        --use_eval
            by default, do not start evaluation. If set`, start evaluation alongside with training.
        --eval_interval <int>
            time duration between contiunous twice evaluation progress.
        --eval_episodes <int>
            number of episodes of a single evaluation.

    Render parameters:
        --save_gifs
            by default, do not save render video. If set, save video.
        --use_render
            by default, do not render the env during training. If set, start render. Note: something, the environment has internal render process which is not controlled by this hyperparam.
        --render_episodes <int>
            the number of episodes to render a given env
        --ifi <float>
            the play interval of each rendered image in saved video.

    Pretrained parameters:
        --model_dir <str>
            by default None. set the path to pretrained model.
    """
    parser = argparse.ArgumentParser(
        description='tartrl', formatter_class=argparse.RawDescriptionHelpFormatter)

    # prepare parameters
    parser.add_argument("--algorithm_name", type=str,
                        default='mappo',
                        choices=["mappo_transformer", "rmappo", "mappo", "rmappg", "mappg", "rqmix", "rvdn", "rmaddpg",
                                 "rmatd3", "rmappo_xzl", "rmappo_new", "transmit"])

    parser.add_argument("--experiment_name", type=str, default="check",
                        help="an identifier to distinguish different experiment.")
    parser.add_argument("--seed", type=int, default=1,
                        help="Random seed for numpy/torch")
    parser.add_argument("--gpu_usage_type", type=str, default='auto', choices=["auto", "single"],
                        help="by default auto, will determine the GPU automatically. If using single, use only use single GPU.")
    parser.add_argument("--disable_cuda", action='store_true', default=False,
                        help="by default False, will use GPU to train; or else will use CPU;")
    parser.add_argument("--cuda_deterministic",
                        action='store_false', default=True,
                        help="by default, make sure random seed effective. if set, bypass such function.")
    parser.add_argument("--pytorch_threads", type=int,
                        default=1, help="Number of torch threads for training")
    parser.add_argument("--n_rollout_threads", type=int, default=32,
                        help="Number of parallel envs for training rollout")
    parser.add_argument("--n_eval_rollout_threads", type=int, default=1,
                        help="Number of parallel envs for evaluating rollout")
    parser.add_argument("--n_render_rollout_threads", type=int, default=1,
                        help="Number of parallel envs for rendering rollout")
    parser.add_argument("--num_env_steps", type=int, default=10e6,
                        help='Number of environment steps to train (default: 10e6)')
    parser.add_argument("--user_name", type=str, default='tartrl',
                        help="user name for the running process")
    parser.add_argument("--wandb_entity", type=str,
                        help="[for wandb usage], to specify entity for simply collecting training data.")
    parser.add_argument("--disable_wandb", action='store_true', default=False,
                        help="[for wandb usage], by default False, will log date to wandb server. or else will use tensorboard to log data.")

    # env parameters
    parser.add_argument("--env_name", type=str, default='StarCraft2',
                        help="specify the name of environment")
    parser.add_argument("--scenario_name", type=str, default='default',
                        help="specify the name of scenario")
    parser.add_argument("--num_agents", type=int,
                        default=2, help="number of players")
    parser.add_argument("--num_enemies", type=int,
                        default=2, help="number of enemies")

    parser.add_argument("--use_obs_instead_of_state", action='store_true',
                        default=False, help="Whether to use global state or concatenated obs")

    # replay buffer parameters
    parser.add_argument("--episode_length", type=int,
                        default=200, help="episode length for training")
    parser.add_argument("--eval_episode_length", type=int,
                        default=200, help="episode length for evaluation")
    parser.add_argument("--max_episode_length", type=int,
                        default=None, help="Max length for any episode")

    # network parameters
    parser.add_argument("--separate_policy", action='store_true',
                        default=False, help='Whether agent seperate the policy')
    parser.add_argument("--concat_obs_as_share_obs", action='store_true',
                        default=False, help="Whether to concatenate all the observations as the share observation")
    parser.add_argument("--use_conv1d", action='store_true',
                        default=False, help="Whether to use conv1d")
    parser.add_argument("--stacked_frames", type=int, default=1,
                        help="Dimension of hidden layers for actor/critic networks")
    parser.add_argument("--use_stacked_frames", action='store_true',
                        default=False, help="Whether to use stacked_frames")
    parser.add_argument("--hidden_size", type=int, default=64,
                        help="Dimension of hidden layers for actor/critic networks")  # different network may need different size
    parser.add_argument("--layer_N", type=int, default=1,
                        help="Number of layers for actor/critic networks")
    parser.add_argument("--activation_id", type=int,
                        default=1, help="choose 0 to use tanh, 1 to use relu, 2 to use leaky relu, 3 to use elu")
    parser.add_argument("--use_popart", action='store_true', default=False,
                        help="by default False, use PopArt to normalize rewards.")
    parser.add_argument("--dual_clip_ppo", action='store_true', default=False,
                        help="by default False, use dual-clip ppo.")
    parser.add_argument("--dual_clip_coeff", type=float, default=3,
                        help="by default 3, use PopArt to normalize rewards.")
    parser.add_argument("--use_valuenorm", action='store_true', default=False,
                        help="by default False, use running mean and std to normalize rewards.")
    parser.add_argument("--use_feature_normalization", action='store_false',
                        default=True, help="Whether to apply layernorm to the inputs")
    parser.add_argument("--use_orthogonal", action='store_false', default=True,
                        help="Whether to use Orthogonal initialization for weights and 0 initialization for biases")
    parser.add_argument("--gain", type=float, default=0.01,
                        help="The gain # of last action layer")
    parser.add_argument("--cnn_layers_params", type=str, default=None,
                        help="The parameters of cnn layer")
    parser.add_argument("--use_maxpool2d", action='store_true',
                        default=False, help="Whether to apply layernorm to the inputs")

    parser.add_argument("--rnn_type", type=str, default="gru", choices=["gru", "lstm"], help="rnn types: gru or lstm")
    parser.add_argument("--rnn_num", type=int, default=1, help="rnn layer number")

    # recurrent parameters
    parser.add_argument("--use_naive_recurrent_policy", action='store_true',
                        default=False, help='Whether to use a naive recurrent policy')
    parser.add_argument("--use_recurrent_policy", action='store_true',
                        default=False, help='use a recurrent policy')
    parser.add_argument("--recurrent_N", type=int, default=1,
                        help="The number of recurrent layers.")
    parser.add_argument("--data_chunk_length", type=int, default=10,
                        help="Time length of chunks used to train a recurrent_policy")
    parser.add_argument("--use_influence_policy", action='store_true',
                        default=False, help='use a recurrent policy')
    parser.add_argument("--influence_layer_N", type=int, default=1,
                        help="Number of layers for actor/critic networks")

    # attn parameters
    parser.add_argument("--use_attn", action='store_true', default=False,
                        help=" by default False, use attention tactics.")
    parser.add_argument("--attn_N", type=int, default=1,
                        help="the number of attn layers, by default 1")
    parser.add_argument("--attn_size", type=int, default=64,
                        help="by default, the hidden size of attn layer")
    parser.add_argument("--attn_heads", type=int, default=4,
                        help="by default, the # of multiply heads")
    parser.add_argument("--dropout", type=float, default=0.0,
                        help="by default 0, the dropout ratio of attn layer.")
    parser.add_argument("--use_average_pool",
                        action='store_false', default=True, help="by default True, use average pooling for attn model.")
    parser.add_argument("--use_attn_internal", action='store_false', default=True,
                        help="by default True, whether to strengthen own characteristics")
    parser.add_argument("--use_cat_self", action='store_false', default=True,
                        help="by default True, whether to strengthen own characteristics")

    # optimizer parameters
    parser.add_argument("--lr", type=float, default=5e-4,
                        help='learning rate (default: 5e-4)')
    parser.add_argument("--tau", type=float, default=0.995,
                        help='soft update polyak (default: 0.995)')
    parser.add_argument("--critic_lr", type=float, default=5e-4,
                        help='critic learning rate (default: 5e-4)')
    parser.add_argument("--opti_eps", type=float, default=1e-5,
                        help='RMSprop optimizer epsilon (default: 1e-5)')
    parser.add_argument("--weight_decay", type=float, default=0,
                        help='weight decay (defaul: 0)')

    # ppo parameters
    parser.add_argument("--ppo_epoch", type=int, default=15,
                        help='number of ppo epochs (default: 15)')
    parser.add_argument("--use_policy_vhead",
                        action='store_true', default=False,
                        help="by default, do not use policy vhead. if set, use policy vhead.")
    parser.add_argument("--use_clipped_value_loss",
                        action='store_false', default=True,
                        help="by default, clip loss value. If set, do not clip loss value.")
    parser.add_argument("--clip_param", type=float, default=0.2,
                        help='ppo clip parameter (default: 0.2)')
    parser.add_argument("--num_mini_batch", type=int, default=1,
                        help='number of batches for ppo (default: 1)')
    parser.add_argument("--policy_value_loss_coef", type=float,
                        default=0.5, help='policy value loss coefficient (default: 0.5)')
    parser.add_argument("--entropy_coef", type=float, default=0.01,
                        help='entropy term coefficient (default: 0.01)')
    parser.add_argument("--value_loss_coef", type=float,
                        default=0.5, help='value loss coefficient (default: 0.5)')
    parser.add_argument("--use_max_grad_norm",
                        action='store_false', default=True,
                        help="by default, use max norm of gradients. If set, do not use.")
    parser.add_argument("--max_grad_norm", type=float, default=10.0,
                        help='max norm of gradients (default: 0.5)')
    parser.add_argument("--use_gae", action='store_false',
                        default=True, help='use generalized advantage estimation')
    parser.add_argument("--gamma", type=float, default=0.99,
                        help='discount factor for rewards (default: 0.99)')
    parser.add_argument("--gae_lambda", type=float, default=0.95,
                        help='gae lambda parameter (default: 0.95)')
    parser.add_argument("--use_proper_time_limits", action='store_true',
                        default=False, help='compute returns taking into account time limits')
    parser.add_argument("--use_huber_loss", action='store_false', default=True,
                        help="by default, use huber loss. If set, do not use huber loss.")
    parser.add_argument("--use_value_active_masks",
                        action='store_false', default=True,
                        help="by default True, whether to mask useless data in value loss.")
    parser.add_argument("--use_policy_active_masks",
                        action='store_false', default=True,
                        help="by default True, whether to mask useless data in policy loss.")
    parser.add_argument("--huber_delta", type=float,
                        default=10.0, help=" coefficience of huber loss.")

    parser.add_argument("--use_adv_normalize", action='store_true', default=False,
                        help="whether to normalize advantage")

    # ppg parameters
    parser.add_argument("--aux_epoch", type=int, default=5,
                        help='number of auxiliary epochs (default: 4)')
    parser.add_argument("--clone_coef", type=float, default=1.0,
                        help='clone term coefficient (default: 0.01)')
    parser.add_argument("--use_single_network", action='store_true',
                        default=False, help="share base network between policy network and value network")

    # run parameters
    parser.add_argument("--use_linear_lr_decay", action='store_true',
                        default=False, help='use a linear schedule on the learning rate')
    # save parameters
    parser.add_argument("--save_interval", type=int, default=1,
                        help="time duration between contiunous twice models saving.")

    parser.add_argument("--only_eval", default=False, action="store_true",
                        help="only execute evaluation, default False.")

    # log parameters
    parser.add_argument("--log_interval", type=int, default=5,
                        help="time duration between contiunous twice log printing.")

    # eval parameters
    parser.add_argument("--use_eval", action='store_true', default=False,
                        help="by default, do not start evaluation. If set`, start evaluation alongside with training.")
    parser.add_argument("--eval_interval", type=int, default=25,
                        help="time duration between contiunous twice evaluation progress.")
    parser.add_argument("--eval_episodes", type=int, default=32,
                        help="number of episodes of the evaluation.")

    # render parameters
    parser.add_argument("--save_gifs", action='store_true', default=False,
                        help="by default, do not save render video. If set, save video.")
    parser.add_argument("--use_render", action='store_true', default=False,
                        help="by default, do not render the env during training. If set, start render. Note: something, the environment has internal render process which is not controlled by this hyperparam.")
    parser.add_argument("--render_episodes", type=int, default=5,
                        help="the number of episodes to render a given env")
    parser.add_argument("--ifi", type=float, default=0.1,
                        help="the play interval of each rendered image in saved video.")

    # pretrained parameters
    parser.add_argument("--model_dir", type=str, default=None,
                        help="by default None. set the path to pretrained model.")
    parser.add_argument("--save_dir", type=str, default=None,
                        help="by default None. set the path to save info.")
    parser.add_argument("--init_dir", type=str, default=None,
                            help="use exist enemy to init model; if init_dir, then don't use model_dir")

    parser.add_argument("--run_dir", type=str, default=None,
                        help="root dir to save curves, logs and models.")

    # replay buffer parameters
    parser.add_argument("--use_transmit", action='store_true', default=False,
                        help="by default, do not use transmit. If set`, use transmit as the replay buffer.")

    # reverb server address
    parser.add_argument("--server_address", type=str, default=None,
                        help="Replay buffer server address.")

    parser.add_argument("--use_tlaunch", action='store_true', default=False,
                        help="whether use tlaunch.")

    parser.add_argument("--actor_num", type=int, default=1,
                        help="number of actors")

    # distributed parameters
    parser.add_argument("--distributed_type", type=str, default="sync",
                        help="distributed type to use actors.", choices=["sync", "async"])
    parser.add_argument("--program_type", type=str, default="local",
                        help="running type of current program.",
                        choices=["local", "whole", "actor", "learner", "server", "server_learner", "local_evaluator",
                                 "remote_evaluator"])

    # replay buffer parameters
    parser.add_argument('--use_reward_normalization', action='store_true',
                        default=False, help="Whether to normalize rewards in replay buffer")
    parser.add_argument('--buffer_size', type=int, default=5000,
                        help="Max # of transitions that replay buffer can contain")
    parser.add_argument('--popart_update_interval_step', type=int, default=2,
                        help="After how many train steps popart should be updated")

    # prioritized experience replay
    parser.add_argument('--use_per', action='store_true', default=False,
                        help="Whether to use prioritized experience replay")
    parser.add_argument('--per_alpha', type=float, default=0.6,
                        help="Alpha term for prioritized experience replay")
    parser.add_argument('--per_beta_start', type=float, default=0.4,
                        help="Starting beta term for prioritized experience replay")
    parser.add_argument('--per_eps', type=float, default=1e-6,
                        help="Eps term for prioritized experience replay")
    parser.add_argument('--per_nu', type=float, default=0.9,
                        help="Weight of max TD error in formation of PER weights")

    # off-policy
    parser.add_argument('--batch_size', type=int, default=32,
                        help="Number of buffer transitions to train on at once")
    parser.add_argument('--actor_train_interval_step', type=int, default=2,
                        help="After how many critic updates actor should be updated")
    parser.add_argument('--train_interval_episode', type=int, default=1,
                        help="Number of env steps between updates to actor/critic")
    parser.add_argument('--train_interval', type=int, default=100,
                        help="Number of episodes between updates to actor/critic")
    parser.add_argument('--use_same_share_obs', action='store_false',
                        default=True, help="whether all agents share the same centralized observation, in mpe")
    parser.add_argument('--use_global_all_local_state', action='store_true',
                        default=False, help="Whether to use available actions, in smac")
    parser.add_argument('--prev_act_inp', action='store_true', default=False,
                        help="Whether the actor input takes in previous actions as part of its input")

    # update parameters
    parser.add_argument('--use_soft_update', action='store_false',
                        default=True, help="Whether to use soft update")
    parser.add_argument('--hard_update_interval_episode', type=int, default=200,
                        help="After how many episodes the lagging target should be updated")

    # exploration parameters
    parser.add_argument('--num_random_episodes', type=int, default=5,
                        help="Number of episodes to add to buffer with purely random actions")
    parser.add_argument('--epsilon_start', type=float, default=1.0,
                        help="Starting value for epsilon, for eps-greedy exploration")
    parser.add_argument('--epsilon_finish', type=float, default=0.05,
                        help="Ending value for epsilon, for eps-greedy exploration")
    parser.add_argument('--epsilon_anneal_time', type=int, default=50000,
                        help="Number of episodes until epsilon reaches epsilon_finish")

    # qmix parameters
    parser.add_argument('--use_double_q', action='store_false',
                        default=True, help="Whether to use double q learning")
    parser.add_argument('--hypernet_layers', type=int, default=2,
                        help="Number of layers for hypernetworks. Must be either 1 or 2")
    parser.add_argument('--mixer_hidden_dim', type=int, default=32,
                        help="Dimension of hidden layer of mixing network")
    parser.add_argument('--hypernet_hidden_dim', type=int, default=64,
                        help="Dimension of hidden layer of hypernetwork (only applicable if hypernet_layers == 2")

    # rmatd3 parameters
    parser.add_argument("--target_action_noise_std", default=0.2,
                        help="Target action smoothing noise for matd3")

    # tlaunch related parameters
    parser.add_argument("--share_temp_dir", default=None, help="temp directory to store job.pkl")
    parser.add_argument("--share_entry_script_path", default=None, help="common path for the process_entry.py file")

    # add for transformer
    parser.add_argument("--encode_state", action='store_true', default=False)
    parser.add_argument("--n_block", type=int, default=1)
    parser.add_argument("--n_embd", type=int, default=64)
    parser.add_argument("--n_head", type=int, default=1)
    parser.add_argument("--dec_actor", action='store_true', default=False)
    parser.add_argument("--share_actor", action='store_true', default=False)

    # hierarchical parameters
    parser.add_argument("--step_difference", type=int, default=1,
                        help="Frequency difference between Controller's step and Executor's step")

    # for GAIL
    parser.add_argument(
        '--gail',
        action='store_true',
        default=False,
        help='do imitation learning with gail')
    parser.add_argument(
        '--gail-experts-dir',
        default='./gail_experts',
        help='directory that contains expert demonstrations for gail')
    parser.add_argument(
        '--gail_batch_size',
        type=int,
        default=128,
        help='gail batch size (default: 128)')
    parser.add_argument(
        '--dis_input_len',
        type=int,
        default=None,
        help='gail input length')
    parser.add_argument(
        '--gail_loss_target',
        type=float,
        default=None,
        help='gail loss target at warm up')

    parser.add_argument(
        '--gail_epoch', type=int, default=5, help='gail epochs (default: 5)')
    parser.add_argument(
        '--disable_action', action='store_true', default=False, help='whether to use action as the input of the '
                                                                     'discriminator')
    parser.add_argument(
        '--gail_hidden_size', type=int, default=256, help='gail hidden state size (default: 256)')
    parser.add_argument(
        '--gail_layer_num', type=int, default=3, help='gail hidden layer number (default: 3)')

    parser.add_argument("--gail_lr", type=float, default=5e-4,
                        help='learning rate (default: 5e-4)')

    # For data collector
    parser.add_argument("--data_dir", type=str, default=None,
                        help="data save directory.")
    parser.add_argument("--force_rewrite", action='store_true', default=False,
                        help="by default False, will delete the data save directory if it exists.")
    parser.add_argument("--collector_num", type=int, default=1,
                        help="number of collectors")

    # For convert
    parser.add_argument("--input_data_dir", type=str, default=None,
                        help="input save directory.")
    parser.add_argument("--output_data_dir", type=str, default=None,
                        help="output data directory.")
    parser.add_argument("--worker_num", type=int, default=1,
                        help="number of workers")
    parser.add_argument(
        '--sample_interval', type=int, default=1, help='data sample interval')

    # For self-play
    parser.add_argument("--self_play", action='store_true', default=False,
                        help="whether to use selfplay")
    parser.add_argument("--selfplay_algo", type=str, default="WeightExistEnemy",
                        help="choose selfplay algorithm")
    parser.add_argument("--max_play_num", type=int, default=2000,
                        help="upper bound of each enemy's play list length")
    parser.add_argument("--max_enemy_num", type=int, default=-1,
                        help="upper bound of enemy model's number exclusive of existing enemies")
    parser.add_argument("--exist_enemy_num", type=int, default=0, help="exist enemy num")

    parser.add_argument("--random_pos", type=int, default=-1, help="random enemy model's position in enemy pool")
    parser.add_argument("--build_in_pos", type=int, default=-1,
                        help="build-in enemy model's position in enemy pool")

    # For joint action
    parser.add_argument("--use_joint_action_loss", action='store_true', default=False,
                        help="whether to use joint action loss")

    # For game wrapper
    parser.add_argument("--frameskip", type=int, default=None, help="whether to use frameskip, default is None")

    # For evaluation
    parser.add_argument("--eval_render", default=False, action='store_true', help="whether to render during evaluation")

    # For JiDi evaluation
    # parser.add_argument("--switch_two_side", default=False, action="store_true",help="whether to evaluate twice to switch two side")

    # For TLaunch
    parser.add_argument("--terminal", default="current_terminal",
                        choices=["local", "current_terminal", "tmux_session", "ssh_tmux_session","k8s","k8s_single"],
                        help="which terminal to use")
    parser.add_argument("--learner_num", type=int, default=1,
                        help="number of learners")
    parser.add_argument("--fetch_num", type=int, default=1,
                        help="number of actors' data to train for a learner")
    parser.add_argument("--tmux_prefix", default=None, type=str, help="prefix which will be added to tmux session")
    parser.add_argument("--kill_all", default=False, action='store_true', help="kill all the tmux session")
    parser.add_argument("--namespace", default="default", type=str, help="namespace of pods")

    # For k8s
    parser.add_argument("--mount_path", default=None, type=str, help="Volume mount path")
    parser.add_argument("--mount_name", default=None, type=str, help="Volume mount name")
    parser.add_argument("--persistent_volume_claim_name", default=None, type=str, help="Persistent volume claim name")

    # For debug
    parser.add_argument('--disable_training', action='store_true', default=False,
                        help="disable training")

    # For Actor
    parser.add_argument('--use_half_actor', action='store_true', default=False,
                        help="whether to use half float for actors")

    # For amp
    parser.add_argument('--use_amp', action='store_true', default=False,
                        help="use mixed precision training")

    # For optimizer
    parser.add_argument('--load_optimizer', action='store_true', default=False,
                        help="whether to restore optimizer")
    return parser



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import torch
import torch.nn as nn

from tartrl.modules.networks.utils.util import init
from tartrl.modules.networks.utils.mlp import MLPBase, MLPLayer
from tartrl.modules.networks.utils.rnn import RNNLayer
from tartrl.modules.networks.utils.act import ACTLayer
from tartrl.modules.networks.utils.popart import PopArt
from tartrl.buffers.utils.util import get_shape_from_obs_space
from tartrl.modules.networks.utils.mix import MIXBase
from tartrl.modules.networks.utils.cnn import CNNBase
from tartrl.modules.networks.base_policy_network import BasePolicyNetwork
from tartrl.utils.util import check_v2 as check


class PolicyNetwork(BasePolicyNetwork):
    def __init__(
            self,
            args,
            input_space,
            action_space,
            device=torch.device("cpu"),
            use_half=False
    ) -> None:
        super(PolicyNetwork, self).__init__(args, device)
        self.hidden_size = args.hidden_size

        self._gain = args.gain
        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self._use_policy_active_masks = args.use_policy_active_masks
        self._use_naive_recurrent_policy = args.use_naive_recurrent_policy
        self._use_recurrent_policy = args.use_recurrent_policy
        self._use_influence_policy = args.use_influence_policy
        self._influence_layer_N = args.influence_layer_N
        self._use_policy_vhead = args.use_policy_vhead
        self._recurrent_N = args.recurrent_N
        self.use_half = use_half
        self.tpdv = dict(dtype=torch.float32, device=device)

        obs_shape = get_shape_from_obs_space(input_space)

        if 'Dict' in obs_shape.__class__.__name__:
            self._mixed_obs = True
            self.base = MIXBase(args, obs_shape, cnn_layers_params=args.cnn_layers_params)
        else:

            self._mixed_obs = False
            self.base = CNNBase(args, obs_shape) if len(obs_shape) == 3 else MLPBase(args, obs_shape,
                                                                                     use_attn_internal=args.use_attn_internal,
                                                                                     use_cat_self=True)

        input_size = self.base.output_size

        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            self.rnn = RNNLayer(input_size, self.hidden_size, self._recurrent_N, self._use_orthogonal,
                                rnn_type=args.rnn_type)
            input_size = self.hidden_size

        if self._use_influence_policy:
            self.mlp = MLPLayer(obs_shape[0], self.hidden_size,
                                self._influence_layer_N, self._use_orthogonal, self._activation_id)
            input_size += self.hidden_size

        self.act = ACTLayer(action_space, input_size, self._use_orthogonal, self._gain)

        if self._use_policy_vhead:
            init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][self._use_orthogonal]

            def init_(m):
                return init(m, init_method, lambda x: nn.init.constant_(x, 0))

            if self._use_popart:
                self.v_out = init_(PopArt(input_size, 1, device=device))
            else:
                self.v_out = init_(nn.Linear(input_size, 1))
        if use_half:
            self.half()
        self.to(device)

    def forward(
            self,
            forward_type,
            *args,
            **kwargs
    ):
        if forward_type == "original":
            return self.forward_original(*args, **kwargs)
        elif forward_type == "eval_actions":
            return self.eval_actions(*args, **kwargs)
        else:
            raise NotImplementedError

    def forward_original(
            self,
            obs,
            rnn_states,
            masks,
            available_actions=None,
            deterministic=False
    ):
        if self._mixed_obs:
            for key in obs.keys():
                obs[key] = check(obs[key], self.use_half, self.tpdv)
                if self.use_half:
                    obs[key].half()
        else:
            obs = check(obs, self.use_half, self.tpdv)
            # if self.use_half:
            #     obs = obs.half()
        rnn_states = check(rnn_states, self.use_half, self.tpdv)
        masks = check(masks, self.use_half, self.tpdv)

        if available_actions is not None:
            available_actions = check(available_actions, self.use_half, self.tpdv)

        actor_features = self.base(obs)

        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            actor_features, rnn_states = self.rnn(actor_features, rnn_states, masks)

        if self._use_influence_policy:
            mlp_obs = self.mlp(obs)
            actor_features = torch.cat([actor_features, mlp_obs], dim=1)

        actions, action_log_probs = self.act(actor_features, available_actions, deterministic)

        return actions, action_log_probs, rnn_states

    def eval_actions(
            self,
            obs,
            rnn_states,
            action,
            masks,
            available_actions=None,
            active_masks=None
    ):
        if self._mixed_obs:
            for key in obs.keys():
                obs[key] = check(obs[key], self.use_half, self.tpdv)
        else:
            obs = check(obs, self.use_half, self.tpdv)

        rnn_states = check(rnn_states, self.use_half, self.tpdv)
        action = check(action, self.use_half, self.tpdv)
        masks = check(masks, self.use_half, self.tpdv)

        if available_actions is not None:
            available_actions = check(available_actions, self.use_half, self.tpdv)

        if active_masks is not None:
            active_masks = check(active_masks, self.use_half, self.tpdv)

        actor_features = self.base(obs)

        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            actor_features, rnn_states = self.rnn(actor_features, rnn_states, masks)

        if self._use_influence_policy:
            mlp_obs = self.mlp(obs)
            actor_features = torch.cat([actor_features, mlp_obs], dim=1)

        action_log_probs, dist_entropy = self.act.evaluate_actions(actor_features, action, available_actions,
                                                                   active_masks=active_masks if self._use_policy_active_masks else None)

        values = self.v_out(actor_features) if self._use_policy_vhead else None

        return action_log_probs, dist_entropy, values

    def get_policy_values(
            self,
            obs,
            rnn_states,
            masks
    ):
        if self._mixed_obs:
            for key in obs.keys():
                obs[key] = check(obs[key], self.use_half, self.tpdv)
        else:
            obs = check(obs).to(**self.tpdv)
        rnn_states = check(rnn_states, self.use_half, self.tpdv)
        masks = check(masks, self.use_half, self.tpdv)

        actor_features = self.base(obs)

        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            actor_features, rnn_states = self.rnn(actor_features, rnn_states, masks)

        if self._use_influence_policy:
            mlp_obs = self.mlp(obs)
            actor_features = torch.cat([actor_features, mlp_obs], dim=1)

        values = self.v_out(actor_features)

        return values



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import gymnasium as gym
from gymnasium import Env
from gymnasium.wrappers import (
    StepAPICompatibility,
    AutoResetWrapper,
)

from tartrl.envs.wrappers import (
    DictWrapper,
    MultiAgentWrapper,
    SingleEnv,
    VecMonitor,
)


def make(
        id: str,
        add_monitor: bool = True,
        *args,
        **kwargs,
) -> Env:
    env = gym.make(id, *args, **kwargs)
    env = MultiAgentWrapper(env)
    env = DictWrapper(env)
    env = AutoResetWrapper(env)
    env = StepAPICompatibility(
        env,
        output_truncation_bool=False
    )
    env = SingleEnv(env)
    if add_monitor:
        env = VecMonitor(env)

    return env



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2022 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import torch.nn as nn

from tartrl.modules.utils.valuenorm import ValueNorm

class BasePolicyNetwork(nn.Module):
    def __init__(self, args, device):
        super(BasePolicyNetwork, self).__init__()
        self.device = device
        self._use_valuenorm = args.use_valuenorm
        self._use_policy_vhead = args.use_policy_vhead

        if self._use_valuenorm:
            if self._use_policy_vhead:
                self.policy_value_normalizer = ValueNorm(1, device = self.device)
        else:
            if self._use_policy_vhead:
                self.policy_value_normalizer = None




import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from .util import init, get_clones
from .attention import Encoder

class MLPLayer(nn.Module):
    def __init__(self, input_dim, hidden_size, layer_N, use_orthogonal, activation_id):
        super(MLPLayer, self).__init__()
        self._layer_N = layer_N

        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        self.fc1 = nn.Sequential(
            init_(nn.Linear(input_dim, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc_h = nn.Sequential(init_(
            nn.Linear(hidden_size, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc2 = get_clones(self.fc_h, self._layer_N)

    def forward(self, x):
        x = self.fc1(x)
        for i in range(self._layer_N):
            x = self.fc2[i](x)
        return x

class CONVLayer(nn.Module):
    def __init__(self, input_dim, hidden_size, use_orthogonal, activation_id):
        super(CONVLayer, self).__init__()

        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        self.conv = nn.Sequential(
                init_(nn.Conv1d(in_channels=input_dim, out_channels=hidden_size//4, kernel_size=3, stride=2, padding=0)), active_func, #nn.BatchNorm1d(hidden_size//4),
                init_(nn.Conv1d(in_channels=hidden_size//4, out_channels=hidden_size//2, kernel_size=3, stride=1, padding=1)), active_func, #nn.BatchNorm1d(hidden_size//2),
                init_(nn.Conv1d(in_channels=hidden_size//2, out_channels=hidden_size, kernel_size=3, stride=1, padding=1)), active_func) #, nn.BatchNorm1d(hidden_size))

    def forward(self, x):
        x = self.conv(x)
        return x


class MLPBase(nn.Module):
    def __init__(self, args, obs_shape, use_attn_internal=False, use_cat_self=True):
        super(MLPBase, self).__init__()

        self._use_feature_normalization = args.use_feature_normalization
        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self._use_attn = args.use_attn
        self._use_attn_internal = use_attn_internal
        self._use_average_pool = args.use_average_pool
        self._use_conv1d = args.use_conv1d
        self._stacked_frames = args.stacked_frames
        self._layer_N = 0 if args.use_single_network else args.layer_N
        self._attn_size = args.attn_size
        self.hidden_size = args.hidden_size

        obs_dim = obs_shape[0]

        if self._use_feature_normalization:
            self.feature_norm = nn.LayerNorm(obs_dim)

        if self._use_attn and self._use_attn_internal:
        
            if self._use_average_pool:
                if use_cat_self:
                    inputs_dim = self._attn_size + obs_shape[-1][1]
                else:
                    inputs_dim = self._attn_size
            else:
                split_inputs_dim = 0
                split_shape = obs_shape[1:]
                for i in range(len(split_shape)):
                    split_inputs_dim += split_shape[i][0]
                inputs_dim = split_inputs_dim * self._attn_size
            self.attn = Encoder(args, obs_shape, use_cat_self)
            self.attn_norm = nn.LayerNorm(inputs_dim)
        else:
            inputs_dim = obs_dim

        if self._use_conv1d:
            self.conv = CONVLayer(self._stacked_frames, self.hidden_size, self._use_orthogonal, self._activation_id)
            random_x = torch.FloatTensor(1, self._stacked_frames, inputs_dim//self._stacked_frames)
            random_out = self.conv(random_x)
            assert len(random_out.shape)==3
            inputs_dim = random_out.size(-1) * random_out.size(-2)

        self.mlp = MLPLayer(inputs_dim, self.hidden_size,
                              self._layer_N, self._use_orthogonal, self._activation_id)

    def forward(self, x):
        if self._use_feature_normalization:
            x = self.feature_norm(x)

        if self._use_attn and self._use_attn_internal:
            x = self.attn(x, self_idx=-1)
            x = self.attn_norm(x)

        if self._use_conv1d:
            batch_size = x.size(0)
            x = x.view(batch_size, self._stacked_frames, -1)
            x = self.conv(x)
            x = x.view(batch_size, -1)

        x = self.mlp(x)

        return x

    @property
    def output_size(self):
        return self.hidden_size


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

from typing import Optional, Union

import numpy as np
import gym
import torch

from tartrl.modules.rl_module import RLModule
from tartrl.modules.model_config import ModelTrainConfig

from tartrl.modules.networks.policy_value_network import PolicyValueNetwork
from tartrl.modules.networks.policy_network import PolicyNetwork
from tartrl.modules.networks.value_network import ValueNetwork
from tartrl.modules.utils.util import update_linear_schedule


class PPOModule(RLModule):
    def __init__(
            self,
            args,
            policy_input_space: gym.spaces.Box,
            critic_input_space: gym.spaces.Box,
            act_space: gym.spaces.Box,
            share_model: bool = False,
            device: Union[str, torch.device] = "cpu",
            rank: Optional[int] = None,
            world_size: Optional[int] = None,
    ):
        model_configs = {}
        if share_model:
            model_configs["model"] = ModelTrainConfig(
                lr=args.lr,
                model=PolicyValueNetwork,
                input_space=policy_input_space,
            )
        else:
            model_configs["policy"] = ModelTrainConfig(
                lr=args.lr,
                model=PolicyNetwork,
                input_space=policy_input_space,
            )
            model_configs["critic"] = ModelTrainConfig(
                lr=args.critic_lr,
                model=ValueNetwork,
                input_space=critic_input_space,
            )

        super(PPOModule, self).__init__(
            args=args,
            model_configs=model_configs,
            act_space=act_space,
            rank=rank,
            world_size=world_size,
            device=device,
        )
        self.share_model = share_model
        self.args = args

    def lr_decay(self, episode, episodes):
        if self.share_model:
            update_linear_schedule(self.optimizers['model'], episode, episodes, self.lr)
        else:
            update_linear_schedule(self.optimizers['actor'], episode, episodes, self.lr)
            update_linear_schedule(self.optimizers['critic'], episode, episodes, self.critic_lr)

    def get_actions(self, share_obs, obs, rnn_states_actor, rnn_states_critic, masks, available_actions=None,
                    deterministic=False):
        if self.share_model:
            raise NotImplementedError
        else:
            actions, action_log_probs, rnn_states_actor = self.models['policy']("original", obs, rnn_states_actor, masks,
                                                                                available_actions, deterministic)

            values, rnn_states_critic = self.models['critic'](share_obs, rnn_states_critic, masks)
        return values, actions, action_log_probs, rnn_states_actor, rnn_states_critic

    def get_values(self, share_obs, rnn_states_critic, masks):
        if self.share_model:
            raise NotImplementedError
        else:
            values, _ = self.models['critic'](share_obs, rnn_states_critic, masks)
        return values

    def evaluate_actions(self, share_obs, obs, rnn_states_actor, rnn_states_critic, action, masks,
                         available_actions=None, active_masks=None, critic_masks_batch=None):
        if critic_masks_batch is None:
            critic_masks_batch = masks

        if self.share_model:
            raise NotImplementedError
        else:
            values, _ = self.models['critic'](share_obs, rnn_states_critic, critic_masks_batch)

            action_log_probs, dist_entropy, policy_values = self.models['policy'](
                "eval_actions",
                obs,
                rnn_states_actor,
                action,
                masks,
                available_actions,
                active_masks,
            )

        return values, action_log_probs, dist_entropy, policy_values

    def act(self, obs, rnn_states_actor, masks, available_actions=None, deterministic=False):
        if self.share_model:
            raise NotImplementedError
        else:
            actions, _, rnn_states_actor = self.models['policy'](
                "original",
                obs,
                rnn_states_actor,
                masks,
                available_actions,
                deterministic,
            )
        return actions, rnn_states_actor

    def get_critic_value_normalizer(self):
        if self.share_model:
            raise NotImplementedError
        else:
            return self.models["critic"].value_normalizer

    @staticmethod
    def init_rnn_states(rollout_num: int,
                             agent_num: int,
                             rnn_layers: int,
                             hidden_size: int
                             ):
        masks = np.ones((rollout_num*agent_num, 1), dtype=np.float32)
        rnn_state = np.zeros((rollout_num*agent_num, rnn_layers, hidden_size))
        return rnn_state, masks



from abc import ABC, abstractmethod
import torch


class BaseAlgorithm(ABC):
    def __init__(self,
                 args,
                 init_module,
                 device=torch.device("cpu")):
        self.all_args = args

        if isinstance(device, str):
            device = torch.device(device)

        self.device = device
        self.tpdv = dict(dtype=torch.float32, device=device)

        self.algo_module = init_module

        self.world_size = self.algo_module.world_size
        self.clip_param = args.clip_param
        self.ppo_epoch = args.ppo_epoch
        self.num_mini_batch = args.num_mini_batch
        self.data_chunk_length = args.data_chunk_length
        self.policy_value_loss_coef = args.policy_value_loss_coef
        self.value_loss_coef = args.value_loss_coef
        self.entropy_coef = args.entropy_coef
        self.max_grad_norm = args.max_grad_norm
        self.huber_delta = args.huber_delta

        self._use_recurrent_policy = args.use_recurrent_policy
        self._use_naive_recurrent = args.use_naive_recurrent_policy
        self._use_max_grad_norm = args.use_max_grad_norm
        self._use_clipped_value_loss = args.use_clipped_value_loss
        self._use_huber_loss = args.use_huber_loss
        self._use_popart = args.use_popart
        self._use_valuenorm = args.use_valuenorm
        self._use_value_active_masks = args.use_value_active_masks
        self._use_policy_active_masks = args.use_policy_active_masks
        self._use_policy_vhead = args.use_policy_vhead
        self._use_new_network = hasattr(
            args, "use_new_network") and args.use_new_network
        self._num_agents = args.num_agents

        self._use_adv_normalize = args.use_adv_normalize

        # for tranformer
        self.dec_actor = args.dec_actor

        self.use_amp = args.use_amp

        self.dual_clip_ppo = args.dual_clip_ppo
        self.dual_clip_coeff = torch.tensor(
            args.dual_clip_coeff).to(
            self.device)

        assert not (self._use_popart and self._use_valuenorm), (
            "self._use_popart and self._use_valuenorm can not be set True simultaneously")

    @abstractmethod
    def train(self, buffer, turn_on=True):
        raise NotImplementedError

    def prep_training(self):
        for model in self.algo_module.models.values():
            model.train()

    def prep_rollout(self):
        for model in self.algo_module.models.values():
            model.eval()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
from typing import Dict, Any
from gymnasium.core import ActType
import numpy as np

from .tartrl_wrapper import nest_expand_dim
from .tartrl_wrapper import TARTRLWrapper

class VecInfo:
    def __init__(self, parallel_env_num, agent_num):
        self.parallel_env_num = parallel_env_num
        self.agent_num = agent_num

        self.infos = []
        self.rewards = []

    def statistics(self) -> Dict[str, Any]:

        rewards = np.array(self.rewards).transpose(2,0,1)
        info_dict = {}
        for i in range(self.agent_num):
            info_dict["agent_{}/episode_reward".format(i)] = rewards[i].sum()
        return info_dict

    def append(self,reward,info):
        assert reward.shape[:2] == (self.parallel_env_num, self.agent_num)
        self.infos.append(info)
        self.rewards.append(reward)


    def reset(self):
        self.infos = []
        self.rewards = []


class BaseVecEnv(TARTRLWrapper):
    def __init__(
            self,
            env,
            parallel_env_num: int
    ) -> None:
        self.parallel_env_num = parallel_env_num
        super().__init__(env)

class VecMonitor(TARTRLWrapper):
    def __init__(self, env: BaseVecEnv):
        super().__init__(env)
        self.vec_info = VecInfo(
            self.env.parallel_env_num,
            self.env.agent_num
        )

    @property
    def use_monitor(self):
        return True

    def step(
            self,
            action: ActType
    ):
        returns = self.env.step(action)

        self.vec_info.append(
            reward = returns[1],
            info = returns[-1]
        )

        return returns

    def statistics(self):
        info_dict = self.vec_info.statistics()
        self.vec_info.reset()
        return info_dict


class SingleEnv(BaseVecEnv):
    def __init__(self, env):
        super().__init__(
            env=env,
            parallel_env_num=1
        )

    def reset(self, *, seed, options=None):

        returns = self.env.reset(seed=seed, options=options)

        if isinstance(returns, tuple):
            if len(returns) == 2:
                # obs, info
                return nest_expand_dim(returns[0]), [returns[1]]
            else:
                raise NotImplementedError("Not support reset return length: {}".format(len(returns)))
        else:
            return nest_expand_dim(returns)

    def step(
            self,
            action: ActType
    ):
        returns = self.env.step(action)

        assert isinstance(returns, tuple), "step return must be tuple, but got: {}".format(type(returns))

        if len(returns) == 4:
            # obs reward done info
            return nest_expand_dim(returns[0]), nest_expand_dim(returns[1]), nest_expand_dim(returns[2]), [returns[3]]
        elif len(returns) == 5:
            # obs reward done truncated, info
            return nest_expand_dim(returns[0]), nest_expand_dim(returns[1]), nest_expand_dim(
                returns[2]), nest_expand_dim(returns[3]), \
                [returns[4]]
        else:
            raise NotImplementedError("Not support step return length: {}".format(len(returns)))



from abc import ABC, abstractmethod


class BaseDriver(ABC):
    @abstractmethod
    def __init__(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def run(self,*args,**kwargs):
        raise NotImplementedError



import torch.nn as nn
class FcEncoder(nn.Module):
    def __init__(self, fc_num, input_size, output_size):
        super(FcEncoder, self).__init__()
        self.first_mlp = nn.Sequential(
                nn.Linear(input_size, output_size), nn.ReLU(), nn.LayerNorm(output_size)
            )
        self.mlp = nn.Sequential()
        for _ in range(fc_num - 1):
            self.mlp.append(nn.Sequential(
                nn.Linear(output_size, output_size), nn.ReLU(), nn.LayerNorm(output_size)
            ))

    def forward(self, x):
        output = self.first_mlp(x)
        return self.mlp(output)


import numpy as np

import torch

def check(input):
    output = torch.from_numpy(input) if type(input) == np.ndarray else input
    return output

def check_v2(input,use_half=False,tpdv = None):
    output = torch.from_numpy(input) if type(input) == np.ndarray else input
    if tpdv:
        output = output.to(**tpdv)
    if use_half:
        output = output.half()
    return output

def _t2n(x):
    return x.detach().cpu().numpy()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2021 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import torch
import torch.nn as nn

from tartrl.modules.networks.utils.cnn import CNNBase
from tartrl.modules.networks.utils.mlp import MLPBase, MLPLayer
from tartrl.modules.networks.utils.rnn import RNNLayer
from tartrl.modules.networks.utils.act import ACTLayer
from tartrl.modules.networks.utils.util import init
from tartrl.modules.networks.utils.popart import PopArt
from tartrl.utils.util import check_v2 as check
from tartrl.buffers.utils.util import get_shape_from_obs_space


class PolicyValueNetwork(nn.Module):
    def __init__(self, args, obs_space, share_obs_space, action_space, device=torch.device("cpu"), use_half=False):
        super(PolicyValueNetwork, self).__init__()
        self._gain = args.gain
        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self._recurrent_N = args.recurrent_N
        self._use_naive_recurrent_policy = args.use_naive_recurrent_policy
        self._use_recurrent_policy = args.use_recurrent_policy
        self._concat_obs_as_share_obs = args.concat_obs_as_share_obs
        self._use_popart = args.use_popart
        self.hidden_size = args.hidden_size
        self.device = device
        self.use_half = use_half
        self.tpdv = dict(dtype=torch.float32, device=device)
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][self._use_orthogonal]

        # obs space
        obs_shape = get_shape_from_obs_space(obs_space)

        self.obs_prep = CNNBase(args, obs_shape) if len(obs_shape) == 3 else MLPBase(args, obs_shape,
                                                                                     use_attn_internal=args.use_attn_internal,
                                                                                     use_cat_self=True)

        # share obs space
        if self._concat_obs_as_share_obs:
            share_obs_shape = get_shape_from_obs_space(share_obs_space)
            self.share_obs_prep = CNNBase(args, share_obs_shape) if len(obs_shape) == 3 else MLPBase(args,
                                                                                                     share_obs_shape,
                                                                                                     use_attn_internal=True,
                                                                                                     use_cat_self=args.use_cat_self)
        else:
            self.share_obs_prep = self.obs_prep

        # common layer
        self.common = MLPLayer(self.hidden_size, self.hidden_size, layer_N=0, use_orthogonal=self._use_orthogonal,
                               activation_id=self._activation_id)

        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            self.rnn = RNNLayer(self.hidden_size, self.hidden_size, self._recurrent_N, self._use_orthogonal)

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0))

        input_size = self.hidden_size

        # value
        if self._use_popart:
            self.v_out = init_(PopArt(input_size, 1, device=device))
        else:
            self.v_out = init_(nn.Linear(input_size, 1))

        # action
        self.act = ACTLayer(action_space, self.hidden_size, self._use_orthogonal, self._gain)
        if use_half:
            self.half()
        self.to(self.device)

    def get_actions(self, obs, rnn_states, masks, available_actions=None, deterministic=False):
        obs = check(obs, self.use_half, self.tpdv)
        rnn_states = check(rnn_states, self.use_half, self.tpdv)
        masks = check(masks, self.use_half, self.tpdv)
        if available_actions is not None:
            available_actions = check(available_actions, self.use_half, self.tpdv)

        x = obs
        x = self.obs_prep(x)

        # common
        actor_features = self.common(x)
        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            actor_features, rnn_states = self.rnn(actor_features, rnn_states, masks)

        actions, action_log_probs = self.act(actor_features, available_actions, deterministic)

        return actions, action_log_probs, rnn_states

    def evaluate_actions(self, obs, rnn_states, action, masks, available_actions, active_masks=None):
        obs = check(obs, self.use_half, self.tpdv)
        rnn_states = check(rnn_states, self.use_half, self.tpdv)
        action = check(action, self.use_half, self.tpdv)
        masks = check(masks, self.use_half, self.tpdv)
        if available_actions is not None:
            available_actions = check(available_actions, self.use_half, self.tpdv)
        if active_masks is not None:
            active_masks = check(active_masks, self.use_half, self.tpdv)

        x = obs
        x = self.obs_prep(x)

        actor_features = self.common(x)
        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            actor_features, rnn_states = self.rnn(actor_features, rnn_states, masks)

        action_log_probs, dist_entropy = self.act.evaluate_actions(actor_features, action, available_actions,
                                                                   active_masks)

        return action_log_probs, dist_entropy

    def get_values(self, share_obs, rnn_states, masks):
        share_obs = check(share_obs, self.use_half, self.tpdv)
        rnn_states = check(rnn_states, self.use_half, self.tpdv)
        masks = check(masks, self.use_half, self.tpdv)

        share_x = share_obs
        share_x = self.share_obs_prep(share_x)

        critic_features = self.common(share_x)
        if self._use_naive_recurrent_policy or self._use_recurrent_policy:
            critic_features, rnn_states = self.rnn(critic_features, rnn_states, masks)

        values = self.v_out(critic_features)

        return values, rnn_states



import math
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from .util import init, get_clones

class Encoder(nn.Module):
    def __init__(self, args, split_shape, cat_self=True):
        super(Encoder, self).__init__()
        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self._attn_N = args.attn_N
        self._attn_size = args.attn_size
        self._attn_heads = args.attn_heads
        self._dropout = args.dropout
        self._use_average_pool = args.use_average_pool
        self._cat_self = cat_self
        if self._cat_self:
            self.embedding = CatSelfEmbedding(
                split_shape[1:], self._attn_size, self._use_orthogonal, self._activation_id)
        else:
            self.embedding = Embedding(
                split_shape[1:], self._attn_size, self._use_orthogonal, self._activation_id)

        self.layers = get_clones(EncoderLayer(
            self._attn_size, self._attn_heads, self._dropout, self._use_orthogonal, self._activation_id), self._attn_N)
        self.norm = nn.LayerNorm(self._attn_size)

    def forward(self, x, self_idx=-1, mask=None):
        x, self_x = self.embedding(x, self_idx)
        for i in range(self._attn_N):
            x = self.layers[i](x, mask)
        x = self.norm(x)
        if self._use_average_pool:
            x = torch.transpose(x, 1, 2)
            x = F.avg_pool1d(x, kernel_size=x.size(-1)).view(x.size(0), -1)
            if self._cat_self:
                x = torch.cat((x, self_x), dim=-1)
        x = x.view(x.size(0), -1)
        return x


# [L,[1,2],[1,2],[1,2]]
def split_obs(obs, split_shape):
    start_idx = 0
    split_obs = []
    for i in range(len(split_shape)):
        split_obs.append(
            obs[:, start_idx:(start_idx+split_shape[i][0]*split_shape[i][1])])
        start_idx += split_shape[i][0]*split_shape[i][1]
    return split_obs


class FeedForward(nn.Module):
    def __init__(self, d_model, d_ff=512, dropout=0.0, use_orthogonal=True, activation_id=1):
        super(FeedForward, self).__init__()
        # We set d_ff as a default to 2048
        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        self.linear_1 = nn.Sequential(
            init_(nn.Linear(d_model, d_ff)), active_func, nn.LayerNorm(d_ff))

        self.dropout = nn.Dropout(dropout)
        self.linear_2 = init_(nn.Linear(d_ff, d_model))

    def forward(self, x):
        x = self.dropout(self.linear_1(x))
        x = self.linear_2(x)
        return x


def ScaledDotProductAttention(q, k, v, d_k, mask=None, dropout=None):
    scores = torch.matmul(q, k.transpose(-2, -1)) / math.sqrt(d_k)
    if mask is not None:
        mask = mask.unsqueeze(1)
        scores = scores.masked_fill(mask == 0, -1e9)
    scores = F.softmax(scores, dim=-1)

    if dropout is not None:
        scores = dropout(scores)

    output = torch.matmul(scores, v)
    return output


class MultiHeadAttention(nn.Module):
    def __init__(self, heads, d_model, dropout=0.0, use_orthogonal=True):
        super(MultiHeadAttention, self).__init__()
        
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0))

        self.d_model = d_model
        self.d_k = d_model // heads
        self.h = heads

        self.q_linear = init_(nn.Linear(d_model, d_model))
        self.v_linear = init_(nn.Linear(d_model, d_model))
        self.k_linear = init_(nn.Linear(d_model, d_model))
        self.dropout = nn.Dropout(dropout)
        self.out = init_(nn.Linear(d_model, d_model))

    def forward(self, q, k, v, mask=None):

        bs = q.size(0)

        # perform linear operation and split into h heads

        k = self.k_linear(k).view(bs, -1, self.h, self.d_k)
        q = self.q_linear(q).view(bs, -1, self.h, self.d_k)
        v = self.v_linear(v).view(bs, -1, self.h, self.d_k)

        # transpose to get dimensions bs * h * sl * d_model

        k = k.transpose(1, 2)
        q = q.transpose(1, 2)
        v = v.transpose(1, 2)
        # calculate attention
        scores = ScaledDotProductAttention(
            q, k, v, self.d_k, mask, self.dropout)

        # concatenate heads and put through final linear layer
        concat = scores.transpose(1, 2).contiguous()\
            .view(bs, -1, self.d_model)

        output = self.out(concat)

        return output


class EncoderLayer(nn.Module):
    def __init__(self, d_model, heads, dropout=0.0, use_orthogonal=True, activation_id=False, d_ff=512, use_FF=False):
        super(EncoderLayer, self).__init__()
        self._use_FF = use_FF
        self.norm_1 = nn.LayerNorm(d_model)
        self.norm_2 = nn.LayerNorm(d_model)
        self.attn = MultiHeadAttention(heads, d_model, dropout, use_orthogonal)
        self.ff = FeedForward(d_model, d_ff, dropout, use_orthogonal, activation_id)
        self.dropout_1 = nn.Dropout(dropout)
        self.dropout_2 = nn.Dropout(dropout)

    def forward(self, x, mask):
        x2 = self.norm_1(x)
        x = x + self.dropout_1(self.attn(x2, x2, x2, mask))
        if self._use_FF:
            x2 = self.norm_2(x)
            x = x + self.dropout_2(self.ff(x2))
        return x


class CatSelfEmbedding(nn.Module):
    def __init__(self, split_shape, d_model, use_orthogonal=True, activation_id=1):
        super(CatSelfEmbedding, self).__init__()
        self.split_shape = split_shape

        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        for i in range(len(split_shape)):
            if i == (len(split_shape)-1):
                setattr(self, 'fc_'+str(i), nn.Sequential(init_(
                    nn.Linear(split_shape[i][1], d_model)), active_func, nn.LayerNorm(d_model)))
            else:
                setattr(self, 'fc_'+str(i), nn.Sequential(init_(nn.Linear(
                    split_shape[i][1]+split_shape[-1][1], d_model)), active_func, nn.LayerNorm(d_model)))

    def forward(self, x, self_idx=-1):
        x = split_obs(x, self.split_shape)
        N = len(x)

        x1 = []
        self_x = x[self_idx]
        for i in range(N-1):
            K = self.split_shape[i][0]
            L = self.split_shape[i][1]
            for j in range(K):
                temp = torch.cat((x[i][:, (L*j):(L*j+L)], self_x), dim=-1)
                exec('x1.append(self.fc_{}(temp))'.format(i))
        temp = x[self_idx]
        exec('x1.append(self.fc_{}(temp))'.format(N-1))

        out = torch.stack(x1, 1)

        return out, self_x


class Embedding(nn.Module):
    def __init__(self, split_shape, d_model, use_orthogonal=True, activation_id=1):
        super(Embedding, self).__init__()
        self.split_shape = split_shape

        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        for i in range(len(split_shape)):
            setattr(self, 'fc_'+str(i), nn.Sequential(init_(
                nn.Linear(split_shape[i][1], d_model)), active_func, nn.LayerNorm(d_model)))

    def forward(self, x, self_idx=None):
        x = split_obs(x, self.split_shape)
        N = len(x)

        x1 = []
        for i in range(N):
            K = self.split_shape[i][0]
            L = self.split_shape[i][1]
            for j in range(K):
                temp = x[i][:, (L*j):(L*j+L)]
                exec('x1.append(self.fc_{}(temp))'.format(i))

        out = torch.stack(x1, 1)

        if self_idx is None:
            return out, None
        else:
            return out, x[self_idx]




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
from typing import Any
import gymnasium as gym
import numpy as np


class TARTRLWrapper(gym.Wrapper):
    @property
    def env_name(self):
        if hasattr(self.env, "env_name"):
            return self.env.env_name
        return self.env.unwrapped.spec.id

    @property
    def agent_num(self):
        if hasattr(self.env, "agent_num"):
            return self.env.agent_num
        else:
            raise NotImplementedError("Not support agent_num")

    @property
    def use_monitor(self):
        return False


class TARTRLObservationWrapper(TARTRLWrapper, gym.ObservationWrapper):
    def __init__(self, env):
        super().__init__(env)


class DictWrapper(TARTRLObservationWrapper):
    def __init__(self, env):
        super().__init__(env)
        need_convert = 'Dict' not in self.env.observation_space.__class__.__name__
        if need_convert:
            self.observation_space = gym.spaces.Dict(
                {
                    "obs": self.env.observation_space,
                }
            )

    def observation(self, observation):
        return {
            "obs": observation,
            "share_obs": observation
        }


class MultiAgentWrapper(TARTRLWrapper):
    def __init__(self, env):
        super().__init__(env)
        self._agent_num = 1

    @property
    def agent_num(self):
        return self._agent_num

    def reset(self, *, seed, options=None):

        returns = self.env.reset(seed=seed, options=options)

        if isinstance(returns, tuple):
            if len(returns) == 2:
                # obs, info
                return nest_expand_dim(returns[0]), returns[1]
            else:
                raise NotImplementedError("Not support reset return length: {}".format(len(returns)))
        else:
            return nest_expand_dim(returns)

    def step(self, action):
        returns = self.env.step(action)

        assert isinstance(returns, tuple), "step return must be tuple, but got: {}".format(type(returns))

        if len(returns) == 4:
            # obs reward done info
            return nest_expand_dim(returns[0]), nest_expand_dim(returns[1]), nest_expand_dim(returns[2]), returns[3]
        elif len(returns) == 5:
            # obs reward done truncated, info
            return nest_expand_dim(returns[0]), nest_expand_dim(returns[1]), nest_expand_dim(
                returns[2]), nest_expand_dim(returns[3]), \
                returns[4]
        else:
            raise NotImplementedError("Not support step return length: {}".format(len(returns)))


def nest_expand_dim(input: Any) -> Any:
    if isinstance(input, (np.ndarray, float, int)):
        return np.expand_dims(input, 0)
    elif isinstance(input, list):
        return [input]
    elif isinstance(input, dict):
        for key in input:
            input[key] = nest_expand_dim(input[key])
        return input
    else:
        raise NotImplementedError("Not support type: {}".format(type(input)))



import numpy as np

import torch
import torch.nn as nn

from .util import init

class Flatten(nn.Module):
    def forward(self, x):
        return x.reshape(x.size(0), -1)

class MIXBase(nn.Module):
    def __init__(self, args, obs_shape, cnn_layers_params=None):
        super(MIXBase, self).__init__()

        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self._use_maxpool2d = args.use_maxpool2d
        self.hidden_size = args.hidden_size
        self.cnn_keys = []
        self.embed_keys = []
        self.mlp_keys = []
        self.n_cnn_input = 0
        self.n_embed_input = 0
        self.n_mlp_input = 0

        for key in obs_shape:
            if obs_shape[key].__class__.__name__ == 'Box':
                key_obs_shape = obs_shape[key].shape
                if len(key_obs_shape) == 3:
                    self.cnn_keys.append(key)
                else:
                    if "orientation" in key:
                        self.embed_keys.append(key)
                    else:
                        self.mlp_keys.append(key)
            else:
                raise NotImplementedError

        if len(self.cnn_keys) > 0:
            self.cnn = self._build_cnn_model(obs_shape, cnn_layers_params, self.hidden_size, self._use_orthogonal, self._activation_id)
        if len(self.embed_keys) > 0:
            self.embed = self._build_embed_model(obs_shape)
        if len(self.mlp_keys) > 0:
            self.mlp = self._build_mlp_model(obs_shape, self.hidden_size, self._use_orthogonal, self._activation_id)

    def forward(self, x):
        out_x = None
        if len(self.cnn_keys) > 0:
            cnn_input = self._build_cnn_input(x)
            out_x = self.cnn(cnn_input)

        if len(self.embed_keys) > 0:
            embed_input = self._build_embed_input(x)
            embed_x = self.embed(embed_input.long()).view(embed_input.size(0), -1)
            if out_x is not None:
                out_x = torch.cat([out_x, embed_x], dim=1)
            else:
                out_x = embed_x
        if len(self.mlp_keys) > 0:
            mlp_input = self._build_mlp_input(x)
            mlp_x = self.mlp(mlp_input).view(mlp_input.size(0), -1)

            if out_x is not None:
                out_x = torch.cat([out_x, mlp_x], dim=1) # ! wrong
            else:
                out_x = mlp_x
        return out_x

    def _build_cnn_model(self, obs_shape, cnn_layers_params, hidden_size, use_orthogonal, activation_id):
        
        if cnn_layers_params is None:
            cnn_layers_params = [(32, 8, 4, 0), (64, 4, 2, 0), (64, 3, 1, 0)]
        else:
            def _convert(params):
                output = []
                for l in params.split(' '):
                    output.append(tuple(map(int, l.split(','))))
                return output
            cnn_layers_params = _convert(cnn_layers_params)
        
        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        for key in self.cnn_keys:
            if key in ['rgb','depth','image','occupy_image']:
                self.n_cnn_input += obs_shape[key].shape[2] 
                cnn_dims = np.array(obs_shape[key].shape[:2], dtype=np.float32)
            elif key in ['global_map','local_map','global_obs','global_merge_obs','global_merge_goal','gt_map']:
                self.n_cnn_input += obs_shape[key].shape[0] 
                cnn_dims = np.array(obs_shape[key].shape[1:3], dtype=np.float32)
            else:
                raise NotImplementedError

        cnn_layers = []
        prev_out_channels = None
        for i, (out_channels, kernel_size, stride, padding) in enumerate(cnn_layers_params):
            if self._use_maxpool2d and i != len(cnn_layers_params) - 1:
                cnn_layers.append(nn.MaxPool2d(2))

            if i == 0:
                in_channels = self.n_cnn_input
            else:
                in_channels = prev_out_channels

            cnn_layers.append(init_(nn.Conv2d(in_channels=in_channels,
                                            out_channels=out_channels,
                                            kernel_size=kernel_size,
                                            stride=stride,
                                            padding=padding,)))
            #if i != len(cnn_layers_params) - 1:
            cnn_layers.append(active_func)
            prev_out_channels = out_channels

        for i, (_, kernel_size, stride, padding) in enumerate(cnn_layers_params):
            if self._use_maxpool2d and i != len(cnn_layers_params) - 1:
                cnn_dims = self._maxpool_output_dim(dimension=cnn_dims,
                dilation=np.array([1, 1], dtype=np.float32),
                kernel_size=np.array([2, 2], dtype=np.float32),
                stride=np.array([2, 2], dtype=np.float32))
            cnn_dims = self._cnn_output_dim(
                dimension=cnn_dims,
                padding=np.array([padding, padding], dtype=np.float32),
                dilation=np.array([1, 1], dtype=np.float32),
                kernel_size=np.array([kernel_size, kernel_size], dtype=np.float32),
                stride=np.array([stride, stride], dtype=np.float32),
            )
            
        cnn_layers += [
            Flatten(),
            init_(nn.Linear(cnn_layers_params[-1][0] * cnn_dims[0] * cnn_dims[1],
                        hidden_size)),
            active_func,
            nn.LayerNorm(hidden_size),
        ]
        return nn.Sequential(*cnn_layers)

    def _build_embed_model(self, obs_shape):
        self.embed_dim = 0
        for key in self.embed_keys:
            self.n_embed_input = 72
            self.n_embed_output = 8
            self.embed_dim += np.prod(obs_shape[key].shape)

        return nn.Embedding(self.n_embed_input, self.n_embed_output)

    def _build_mlp_model(self, obs_shape, hidden_size, use_orthogonal, activation_id):

        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        for key in self.mlp_keys:
            self.n_mlp_input += np.prod(obs_shape[key].shape)

        return nn.Sequential(
                    init_(nn.Linear(self.n_mlp_input, hidden_size)), active_func, nn.LayerNorm(hidden_size))
         
    def _maxpool_output_dim(self, dimension, dilation, kernel_size, stride):
        """Calculates the output height and width based on the input
        height and width to the convolution layer.
        ref: https://pytorch.org/docs/master/nn.html#torch.nn.Conv2d
        """
        assert len(dimension) == 2
        out_dimension = []
        for i in range(len(dimension)):
            out_dimension.append(
                int(np.floor(
                    ((dimension[i] - dilation[i] * (kernel_size[i] - 1) - 1) / stride[i]) + 1
                ))
            )
        return tuple(out_dimension)

    def _cnn_output_dim(self, dimension, padding, dilation, kernel_size, stride):
        """Calculates the output height and width based on the input
        height and width to the convolution layer.
        ref: https://pytorch.org/docs/master/nn.html#torch.nn.Conv2d
        """
        assert len(dimension) == 2
        out_dimension = []
        for i in range(len(dimension)):
            out_dimension.append(
                int(np.floor(
                    ((dimension[i] + 2 * padding[i] - dilation[i] * (kernel_size[i] - 1) - 1) / stride[i]) + 1
                ))
            )
        return tuple(out_dimension)

    def _build_cnn_input(self, obs):
        cnn_input = []

        for key in self.cnn_keys:
            if key in ['rgb','depth','image','occupy_image']:
                cnn_input.append(obs[key].permute(0, 3, 1, 2) / 255.0)
            elif key in ['global_map', 'local_map', 'global_obs', 'global_merge_obs', 'global_merge_goal','gt_map']:
                cnn_input.append(obs[key])
            else:
                raise NotImplementedError

        cnn_input = torch.cat(cnn_input, dim=1)
        return cnn_input

    def _build_embed_input(self, obs):
        embed_input = []
        for key in self.embed_keys:
            embed_input.append(obs[key].view(obs[key].size(0), -1))

        embed_input = torch.cat(embed_input, dim=1)
        return embed_input

    def _build_mlp_input(self, obs):
        mlp_input = []
        for key in self.mlp_keys:
            mlp_input.append(obs[key].view(obs[key].size(0), -1))

        mlp_input = torch.cat(mlp_input, dim=1)
        return mlp_input

    @property
    def output_size(self):
        output_size = 0
        if len(self.cnn_keys) > 0:
            output_size += self.hidden_size

        if len(self.embed_keys) > 0:
            output_size += 8 * self.embed_dim

        if len(self.mlp_keys) > 0:
            output_size += self.hidden_size
        return output_size


import copy

import torch.nn as nn

def init(module, weight_init, bias_init, gain=1):
    weight_init(module.weight.data, gain=gain)
    if module.bias is not None:
        bias_init(module.bias.data)
    return module

def get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])







import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from .util import init

class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)

class CNNLayer(nn.Module):
    def __init__(self, obs_shape, hidden_size, use_orthogonal, activation_id, kernel_size=3, stride=1):
        super(CNNLayer, self).__init__()

        active_func = [nn.Tanh(), nn.ReLU(), nn.LeakyReLU(), nn.ELU()][activation_id]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu', 'leaky_relu', 'leaky_relu'][activation_id])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        input_channel = obs_shape[0]
        input_width = obs_shape[1]
        input_height = obs_shape[2]

        # self.cnn = nn.Sequential(
        #     init_(nn.Conv2d(in_channels=input_channel, out_channels=hidden_size//2, kernel_size=kernel_size, stride=stride)), active_func,
        #     Flatten(),
        #     init_(nn.Linear(hidden_size//2 * (input_width-kernel_size+stride) * (input_height-kernel_size+stride), hidden_size)), active_func,
        #     init_(nn.Linear(hidden_size, hidden_size)), active_func)
        # only for atari
        self.cnn = nn.Sequential(
            init_(nn.Conv2d(input_channel, 32, 8, stride=4)), nn.ReLU(),
            init_(nn.Conv2d(32, 64, 4, stride=2)), nn.ReLU(),
            init_(nn.Conv2d(64, 32, 3, stride=1)), nn.ReLU(), Flatten(),
            init_(nn.Linear(32 * 7 * 7, hidden_size)), nn.ReLU())

    def forward(self, x):
        x = x / 255.0
        x = self.cnn(x)
        
        return x

class CNNBase(nn.Module):
    def __init__(self, args, obs_shape):
        super(CNNBase, self).__init__()

        self._use_orthogonal = args.use_orthogonal
        self._activation_id = args.activation_id
        self.hidden_size = args.hidden_size

        self.cnn = CNNLayer(obs_shape, self.hidden_size, self._use_orthogonal, self._activation_id)

    def forward(self, x):
        x = self.cnn(x)
        return x

    @property
    def output_size(self):
        return self.hidden_size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import numpy as np
from treevalue import TreeValue, reduce_

class ObsData(TreeValue):
    def flatten(self):
        return reduce_(self, lambda **kwargs: np.concatenate(list(kwargs.values())))

    def step_batch(self,step):
        return_dict = {}
        for str_key in self.keys():
            return_dict[str_key] = np.concatenate(self[str_key][step])
        return return_dict

    def __getitem__(self, key):

        if isinstance(key, int):
            return self.step_batch(key)
        else:
            return super().__getitem__(key)

    def step_flatten(self, step):
        reduce_(self, lambda **kwargs: np.concatenate([value[step] for value in kwargs.values()], -1))

def test_basic():
    obs = ObsData({
        'a': np.array([1, 2]),
        'b': np.array([3, 4]),
        'c': np.array([5, 6, 7])
    })

    print(obs['c'])
    obs['d'] = [10]
    print(obs['d'])

    obs['a'][0] = 99
    print(obs['a'])

    print(obs.flatten())
def test_step():
    obs_stepes = [{"obs_a":np.array([[[0,1]]]),"obs_b":np.array([[[3,9]]])},
   {"obs_a":np.array([[[2,4]]]),"obs_b":np.array([[[6,8]]])}]
    obs = ObsData({"obs_a": np.zeros((2,1,1,2)),"obs_b": np.zeros((2,1,1,2))})
    for step in range(len(obs_stepes)):
        for key in obs.keys():
            obs[key][step] = obs_stepes[step][key]
    print(obs[0])
    print(obs[1])


if __name__ == '__main__':
    test_step()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
from abc import ABC, abstractmethod

import torch
from torch.nn.parallel import DistributedDataParallel as DDP

class BaseModule(ABC):
    def __init__(self, args) -> None:
        self.args = args
        self.models = {}
        self.optimizers = {}

    @abstractmethod
    def lr_decay(self, episode: int, episodes: int) -> None:
        raise NotImplementedError

    @abstractmethod
    def restore(self, model_dir: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def save(self, save_dir: str) -> None:
        raise NotImplementedError

    def convert_distributed_model(self) -> None:
        for model_name in self.models:
            model = self.models[model_name]
            model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)
            model = DDP(model, device_ids=[self.device], find_unused_parameters=True)
            self.models[model_name] = model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

from typing import Optional
import torch
import gym


class ModelConfig(dict):
    def __init__(self, *args, **kwargs) -> None:
        super(ModelConfig, self).__init__(*args, **kwargs)


class ModelTrainConfig(ModelConfig):
    def __init__(
            self,
            model: torch.nn.Module,
            input_space: gym.spaces.Box,
            lr: Optional[float] = None,
            *args,
            **kwargs
    ) -> None:
        super(ModelTrainConfig, self).__init__(*args, **kwargs)
        self["model"] = model
        self["input_space"] = input_space
        self["lr"] = lr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import click
from click.core import Context, Option

from tartrl import __TITLE__, __VERSION__, __AUTHOR__, __EMAIL__
from termcolor import colored

def red(text: str):
    return colored(text, 'red')

def print_version(
        ctx: Context,
        param: Option,
        value: bool,
) -> None:
    if not value or ctx.resilient_parsing:
        return
    click.secho(f"{__TITLE__.upper()} version: {red(__VERSION__)}")
    click.secho(f"Developed by {__AUTHOR__}, Email: {red(__EMAIL__)}")
    ctx.exit()


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    '--version',
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Show package's version information."
)
@click.option(
    '--mode',
    prompt='Choose execution mode',
    type=click.Choice(
        [
            'train',
        ]
    ),
    help='execution mode'
)
@click.option('--env', prompt='Please enter environment name', type=str, help='RL environment name')
def run(
        mode: str,
        env: str
):
    if mode == "train":
        from tartrl.cli.train import train_agent
        train_agent(env)
    else:
        raise NotImplementedError



import torch
from torch.distributions import Categorical, Normal
from torch.nn import functional as F


def discrete_autoregreesive_act(decoder, obs_rep, obs, batch_size, n_agent, action_dim, tpdv,
                                available_actions=None, deterministic=False):
    shifted_action = torch.zeros((batch_size, n_agent, action_dim + 1)).to(**tpdv)
    shifted_action[:, 0, 0] = 1
    output_action = torch.zeros((batch_size, n_agent, 1), dtype=torch.long)
    output_action_log = torch.zeros_like(output_action, dtype=torch.float32)

    for i in range(n_agent):
        logit = decoder(shifted_action, obs_rep, obs)[:, i, :]
        if available_actions is not None:
            logit[available_actions[:, i, :] == 0] = -1e10

        distri = Categorical(logits=logit)
        action = distri.probs.argmax(dim=-1) if deterministic else distri.sample()
        action_log = distri.log_prob(action)

        output_action[:, i, :] = action.unsqueeze(-1)
        output_action_log[:, i, :] = action_log.unsqueeze(-1)
        if i + 1 < n_agent:
            shifted_action[:, i + 1, 1:] = F.one_hot(action, num_classes=action_dim)
    return output_action, output_action_log


def discrete_parallel_act(decoder, obs_rep, obs, action, batch_size, n_agent, action_dim, tpdv,
                          available_actions=None):
    one_hot_action = F.one_hot(action.squeeze(-1), num_classes=action_dim)  # (batch, n_agent, action_dim)
    shifted_action = torch.zeros((batch_size, n_agent, action_dim + 1)).to(**tpdv)
    shifted_action[:, 0, 0] = 1
    shifted_action[:, 1:, 1:] = one_hot_action[:, :-1, :]
    logit = decoder(shifted_action, obs_rep, obs)
    if available_actions is not None:
        logit[available_actions == 0] = -1e10

    distri = Categorical(logits=logit)
    action_log = distri.log_prob(action.squeeze(-1)).unsqueeze(-1)
    entropy = distri.entropy().unsqueeze(-1)
    return action_log, entropy


def continuous_autoregreesive_act(decoder, obs_rep, obs, batch_size, n_agent, action_dim, tpdv,
                                  deterministic=False):
    shifted_action = torch.zeros((batch_size, n_agent, action_dim)).to(**tpdv)
    output_action = torch.zeros((batch_size, n_agent, action_dim), dtype=torch.float32)
    output_action_log = torch.zeros_like(output_action, dtype=torch.float32)

    for i in range(n_agent):
        act_mean = decoder(shifted_action, obs_rep, obs)[:, i, :]
        action_std = torch.sigmoid(decoder.log_std) * 0.5

        distri = Normal(act_mean, action_std)
        action = act_mean if deterministic else distri.sample()
        action_log = distri.log_prob(action)

        output_action[:, i, :] = action
        output_action_log[:, i, :] = action_log
        if i + 1 < n_agent:
            shifted_action[:, i + 1, :] = action

    return output_action, output_action_log


def continuous_parallel_act(decoder, obs_rep, obs, action, batch_size, n_agent, action_dim, tpdv):
    shifted_action = torch.zeros((batch_size, n_agent, action_dim)).to(**tpdv)
    shifted_action[:, 1:, :] = action[:, :-1, :]

    act_mean = decoder(shifted_action, obs_rep, obs)
    action_std = torch.sigmoid(decoder.log_std) * 0.5
    distri = Normal(act_mean, action_std)

    action_log = distri.log_prob(action)
    entropy = distri.entropy()
    return action_log, entropy



def get_shape_from_obs_space(obs_space):
    if obs_space.__class__.__name__ == 'Box':
        obs_shape = obs_space.shape
    elif obs_space.__class__.__name__ == 'list':
        obs_shape = obs_space
    elif obs_space.__class__.__name__ == 'Dict':
        obs_shape = obs_space.spaces
    else:
        raise NotImplementedError
    return obs_shape

def get_shape_from_act_space(act_space):
    if act_space.__class__.__name__ == 'Discrete':
        act_shape = 1
    elif act_space.__class__.__name__ == "MultiDiscrete":
        act_shape = act_space.shape
    elif act_space.__class__.__name__ == "Box":
        act_shape = act_space.shape[0]
    elif act_space.__class__.__name__ == "MultiBinary":
        act_shape = act_space.shape[0]
    else:  # agar
        act_shape = act_space[0].shape[0] + 1
    return act_shape



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""

import socket
from typing import Optional, Dict, Any
import os
import logging
from rich.logging import RichHandler
import wandb
from pathlib import Path
import numpy as np
import torch

class Logger:
    def __init__(
            self,
            args,
            project_name: str,
            scenario_name: str,
            wandb_entity: str,
            exp_name: str,
            log_path: Optional[str] = None,
            use_wandb: bool = False,
            use_tensorboard: bool = False,
            log_level: int = logging.DEBUG,
    ) -> None:
        self.use_wandb = use_wandb
        self.use_tensorboard = use_tensorboard

        self.log_level = log_level
        self.log_path = log_path
        self.project_name = project_name
        self.scenario_name = scenario_name
        self.wandb_entity = wandb_entity
        self.exp_name = exp_name
        self.args = args
        self._init()

    def _init(self) -> None:
        running_programs = ["learner", "server_learner", "local", "whole", "local_evaluator"]
        if not self.args.program_type in running_programs:
            return None

        if self.log_path is None:
            run_dir = Path(os.path.split(os.path.dirname(os.path.abspath(__file__)))[
                               0] + "/results") / self.project_name / self.scenario_name / self.args.algorithm_name / self.exp_name
        else:
            run_dir = Path(
                self.log_path) / self.project_name / self.scenario_name / self.args.algorithm_name / self.exp_name

        if not run_dir.exists():
            os.makedirs(str(run_dir))

        if not self.use_wandb:
            if not run_dir.exists():
                curr_run = 'run1'
            else:
                exst_run_nums = [int(str(folder.name).split('run')[1]) for folder in run_dir.iterdir() if
                                 str(folder.name).startswith('run')]
                if len(exst_run_nums) == 0:
                    curr_run = 'run1'
                else:
                    curr_run = 'run%i' % (max(exst_run_nums) + 1)
            run_dir = run_dir / curr_run
            if not run_dir.exists():
                os.makedirs(str(run_dir))

        if hasattr(self.args, "render"):
            self.args.render_save_path = run_dir / "render.png"

        log_path = os.path.join(run_dir, "log.txt")
        logging.basicConfig(
            level=self.log_level,
            format="%(asctime)s [%(levelname)s] %(message)s",
            handlers=[
                logging.FileHandler(log_path),
                RichHandler()
            ]
        )

        if self.use_wandb:
            wandb.init(config=self.args,
                       project=self.project_name,
                       entity=self.wandb_entity,
                       notes=socket.gethostname(),
                       name=str(self.args.algorithm_name) + "_" +
                            str(self.exp_name) +
                            "_seed" + str(self.args.seed),
                       dir=str(run_dir),
                       job_type="training",
                       reinit=True)
        elif self.use_tensorboard:
            from tensorboardX import SummaryWriter
            self.log_dir = str(run_dir / 'logs')
            if not os.path.exists(self.log_dir):
                os.makedirs(self.log_dir)
            self.writter = SummaryWriter(self.log_dir)

        self.run_dir = run_dir

    def close(self):
        if self.use_wandb:
            wandb.finish()

    def info(self, msg: str):
        logging.info(msg)

    def log_learner_info(
            self,
            leaner_id: int,
            infos: Dict[str, Any],
            step: int,
    ) -> None:
        for k, v in infos.items():
            if self.use_wandb:
                wandb.log({"Learner_{}/{}".format(leaner_id, k): v},
                          step=step)
            elif self.use_tensorboard:
                self.writter.add_scalars("Learner_{}/{}".format(leaner_id, k),
                                         {"Learner_{}/{}".format(leaner_id, k): v}, step)


    def log_info(
            self,
            infos: Dict[str, Any],
            step: int,
    ) -> None:
        for k, v in infos.items():
            # print(k,v)

            if isinstance(v,torch.Tensor):
                v = v.item()

            if not isinstance(v, (int,float)):
                v = np.mean(v)

            if self.use_wandb:
                wandb.log({k: v}, step=step)
            elif self.use_tensorboard:
                self.writter.add_scalars(
                    k, {k: v}, step)


if __name__ == "__main__":
    from collections import namedtuple

    ARG = namedtuple("args", ["seed", "algorithm_name", "program_type"])

    class Namespace(ARG):
        @property
        def __dict__(self):
            return self._asdict()


    args = Namespace(seed=1, algorithm_name="algorithm", program_type="local")

    logger = Logger(
        args=args,
        project_name="test_logger",
        scenario_name="logger",
        wandb_entity="tmarl",
        exp_name="test",
        log_path="../../../test_logger/",
        use_wandb=False,
        use_tensorboard=True,
    )
    for step in range(100):
        logger.log_info({"test": step}, step)
        logger.log_info({"test2": [step,2*step]}, step)
        logger.log_info({"test3": np.random.random((10,10))}, step)
    logger.info("hello")
    logger.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2023 The TARTRL Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""""""
import io
from typing import Union, Type, TypeVar
from abc import ABC, abstractmethod
import pathlib

SelfAgent = TypeVar("SelfAgent", bound="BaseAgent")

SelfBaseAgent = TypeVar("SelfBaseAgent", bound="BaseAgent")

class BaseAgent(ABC):
    @abstractmethod
    def __init__(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def save(self, path: Union[str, pathlib.Path, io.BufferedIOBase], ) -> None:
        raise NotImplementedError

    @classmethod
    def load(
            cls: Type[SelfBaseAgent],
            path: Union[str, pathlib.Path, io.BufferedIOBase],
    ) -> SelfBaseAgent:
        agent = cls()
        return agent
